@@ -1,6 +1,6 @@
-// ----------------------------------------------------------------------------
+// -------------------------------------------------------------------------
 //
-//  Copyright (C) 2006-2010 Fons Adriaensen <fons@kokkinizita.net>
+//  Copyright (C) 2006-2011 Fons Adriaensen <fons@linuxaudio.org>
 //    
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -16,35 +16,92 @@
 //  along with this program; if not, write to the Free Software
 //  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 //
-// ----------------------------------------------------------------------------
+// -------------------------------------------------------------------------
 
 
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
+#include <math.h>
 #include <time.h>
 #include "adconf.h"
+#include "sstring.h"
 
 
-static const char *shdr_a  = "#            id      dist     azim     elev     conn";
-static const char *shdr_b  = "#-----------------------------------------------------------------------";
-static const char *mhdr10a = "#            W        X        Y";
-static const char *mhdr10b = "#-----------------------------------";
-static const char *mhdr11a = "#            W        X        Y        Z";
-static const char *mhdr11b = "#--------------------------------------------";
-static const char *mhdr20a = "#            W        X        Y        U        V";
-static const char *mhdr20b = "#-----------------------------------------------------";
-static const char *mhdr21a = "#            W        X        Y        Z        U        V";
-static const char *mhdr21b = "#--------------------------------------------------------------";
-static const char *mhdr22a = "#            W        X        Y        Z        R        S        T        U        V";
-static const char *mhdr22b = "#-----------------------------------------------------------------------------------------";
-static const char *mhdr30a = "#            W        X        Y        U        V        P        Q";
-static const char *mhdr30b = "#-----------------------------------------------------------------------";
-static const char *mhdr31a = "#            W        X        Y        Z        U        V        P        Q";
-static const char *mhdr31b = "#--------------------------------------------------------------------------------";
 static const char *jackp   = "system:playback";
 
+static const int chmap10 [] = { 0, 3, 1, -1 };
+static const int chmap11 [] = { 0, 3, 1, 2, -1 };
+static const int chmap20 [] = { 0, 3, 1, 8, 4, -1 };
+static const int chmap21 [] = { 0, 3, 1, 2, 8, 4, -1 };
+static const int chmap22 [] = { 0, 3, 1, 2, 6, 7, 5, 8, 4, -1 };
+static const int chmap30 [] = { 0, 3, 1, 8, 4, 15, 9, -1 };
+static const int chmap31 [] = { 0, 3, 1, 2, 8, 4, 15, 9, -1 };
+
+
+static const float gc_n3d [16] =
+{
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f,
+    1.0f
+};
+
+
+static const float gc_sn3d [16] =
+{
+    1.0f,
+    1.732051f,
+    1.732051f,
+    1.732051f,
+    2.236068f, 
+    2.236068f, 
+    2.236068f, 
+    2.236068f, 
+    2.236068f, 
+    2.645751f,
+    2.645751f,
+    2.645751f,
+    2.645751f,
+    2.645751f,
+    2.645751f,
+    2.645751f
+};
+
+
+static const float gc_fuma [16] =
+{
+    1.414214f,
+    1.732051f,
+    1.732051f,
+    1.732051f,
+    1.936492f,
+    1.936492f,
+    2.236068f,
+    1.936492f,
+    1.936492f,
+    2.091650f,
+    1.972026f,
+    2.231093f,
+    2.645751f,
+    2.231093f,
+    1.972026f,
+    2.091650f,
+};
+
 
 Speaker::Speaker (void)
 {
@@ -65,6 +122,17 @@ void Speaker::reset (void)
 }
 
 
+void Speaker::rae2xyz (void)
+{
+    float a = RAD(_a);
+    float e = RAD(_e);
+    _x = _r * cosf (a) * cosf (e);
+    _y = _r * sinf (a) * cosf (e);
+    _z = _r * sinf (e);
+}
+
+
+
 Matrow::Matrow (void)
 {
 }
@@ -72,10 +140,68 @@ Matrow::Matrow (void)
 
 void Matrow::reset (void)
 {
-    _w = _x = _y = _z = _r = _s = _t = _u = _v = _p = _q = 0;
+    memset (_coeff, 0, 16 * sizeof (float));
 }
 
 
+void Matrow::check (void)
+{
+    int   i;
+    float s1, s2, s3;
+
+    s1 = s2 = s3 = 0;
+    for (i = 1; i <  4; i++) s1 += _coeff [i] * _coeff [i];
+    for (i = 4; i <  9; i++) s2 += _coeff [i] * _coeff [i];
+    for (i = 9; i < 16; i++) s3 += _coeff [i] * _coeff [i];
+    printf ("%10.6lf %10.6lf %10.6lf %10.6lf\n", _coeff [0], s1, s2, s3);
+}
+
+
+void Matrow::encode (int cm, float x, float y, float z)
+{
+    float x2, y2, z2, c2, s2;
+
+    reset ();
+    if (cm & 1) _coeff [ACN_W] = 1.0f;
+    if (cm & 2) _coeff [ACN_Y] = sqrtf (3.0f) * y;
+    if (cm & 4) _coeff [ACN_Z] = sqrtf (3.0f) * z;
+    if (cm & 8) _coeff [ACN_X] = sqrtf (3.0f) * x;
+    cm >>= 4;
+    if (! cm) return;
+    x2 = x * x;
+    y2 = y * y;
+    z2 = z * z;
+    c2 = (x2 - y2) / 2;
+    s2 = x * y;
+    if (cm &  1) _coeff [ACN_V] = sqrtf (15.0f) * s2;
+    if (cm &  2) _coeff [ACN_T] = sqrtf (15.0f) * y * z;
+    if (cm &  4) _coeff [ACN_R] = sqrtf (5.0f) * (1.5f * z2 - 0.5f);
+    if (cm &  8) _coeff [ACN_S] = sqrtf (15.0f) * x * z;
+    if (cm & 16) _coeff [ACN_U] = sqrtf (15.0f) * c2;
+    cm >>= 5;
+    if (! cm) return;
+    if (cm &  1) _coeff [ACN_Q] = sqrtf (35.0f / 8)  * (3 * x2 - y2) * y;
+    if (cm &  2) _coeff [ACN_O] = sqrtf (105.0f) * s2 * z;
+    if (cm &  4) _coeff [ACN_M] = sqrtf (21.0f / 8)  * (5 * z2 - 1) * y; 
+    if (cm &  8) _coeff [ACN_K] = sqrtf (7.0f / 4)   * (5 * z2 - 3) * z;
+    if (cm & 16) _coeff [ACN_L] = sqrtf (21.0f / 8)  * (5 * z2 - 1) * x; 
+    if (cm & 32) _coeff [ACN_N] = sqrtf (105.0f) * c2 * z;
+    if (cm & 64) _coeff [ACN_P] = sqrtf (35.0f / 8)  * (x2 - 3 * y2) * x;
+}
+
+
+void Matrow::encode (int cm, float az, float el)
+{
+    float x, y, z;
+
+    x = cosf (el) * cosf (az);
+    y = cosf (el) * sinf (az);
+    z = sinf (el);
+    encode (cm, x, y, z);
+}
+
+
+
 AD_conf::AD_conf (void)
 {
     _state     = 0;
@@ -88,10 +214,9 @@ void AD_conf::reset (void)
 {
     int i;
 
-    _fform = 2;
+    _fform = 3;
     _descr [0] = 0;
-    _dec.h_ord = 1;
-    _dec.v_ord = 0;
+    _dec.cmask = 0;
     _dec.nband = 1;
     _dec.nspkr = 4;
     _dec.scale = FUMA;
@@ -112,6 +237,9 @@ void AD_conf::reset (void)
 	_lfmatrix [i].reset ();
 	_hfmatrix [i].reset ();
     }
+    _h_ord = 0;
+    _v_ord = 0;
+    _chmap = 0;
 }
 
 
@@ -132,13 +260,13 @@ int AD_conf::save (const char *file)
         return 1;
     } 
 
+    _fform = 3;
     t = time (0);
     fprintf (F, "# AmbDec configuration\n");
-    fprintf (F, "# Written by AmbDec-%s at %s\n", VERSION, ctime (&t));
+    fprintf (F, "# Written by %s-%s at %s\n", PROGNAME, VERSION, ctime (&t));
     fprintf (F, "/description      %s\n\n", _descr);
     fprintf (F, "/version          %d\n\n", _fform);
-    fprintf (F, "/dec/hor_order    %d\n", _dec.h_ord); 
-    fprintf (F, "/dec/ver_order    %d\n", _dec.v_ord); 
+    fprintf (F, "/dec/chan_mask    %x\n", _dec.cmask); 
     fprintf (F, "/dec/freq_bands   %d\n", _dec.nband); 
     fprintf (F, "/dec/speakers     %d\n", _dec.nspkr); 
     switch (_dec.scale)
@@ -174,7 +302,6 @@ int AD_conf::save (const char *file)
     fprintf (F, "/opt/xover_ratio  %4.1lf\n", _opt.ratio); 
 
     fprintf (F, "\n/speakers/{\n"); 
-    fprintf (F, "%s\n%s\n", shdr_a, shdr_b); 
     for (i = 0, S = _speakers; i < _dec.nspkr; i++, S++)
     {
 	fprintf (F, "add_spkr   %3s  %8.3lf %8.1lf %8.1lf    %s\n", S->_label, S->_r, S->_a, S->_e, S->_jackp);  
@@ -196,83 +323,37 @@ int AD_conf::save (const char *file)
 
 void AD_conf::save_matrix (FILE *F, const char *pref, Matrow *matr, float *gain)
 {
-    int i;
+    int           i, j, m;
+    Matrow        *M;
+    const float   *gc;
  
-    fprintf (F, "/%s/{\n", pref);
-
-    switch (_dec.h_ord)
+    switch (_dec.scale)
     {
-    case 1:
-        fprintf (F, "order_gain    %8.5lf %8.5lf\n", gain [0], gain [1]);
+    case N3D:
+	gc = gc_n3d;
 	break;
-    case 2:
-        fprintf (F, "order_gain    %8.5lf %8.5lf %8.5lf\n", gain [0], gain [1], gain [2]);
+    case SN3D:
+	gc = gc_sn3d;
 	break;
-    case 3:
-        fprintf (F, "order_gain    %8.5lf %8.5lf %8.5lf %8.5lf\n", gain [0], gain [1], gain [2], gain [3]);
-	break;
-    }	
-
-    switch ((_dec.h_ord << 4) | _dec.v_ord)
-    {
-    case 0x10:
-	fprintf (F, "%s\n%s\n", mhdr10a, mhdr10b);
-	for (i = 0; i < _dec.nspkr; i++, matr++)
-	{
-	    fprintf (F, "add_row   %8.5lf %8.5lf %8.5lf\n",
-		     matr->_w, matr->_x, matr->_y); 
-	}
-	break;
-    case 0x11:
-	fprintf (F, "%s\n%s\n", mhdr11a, mhdr11b);
-	for (i = 0; i < _dec.nspkr; i++, matr++)
-	{
-	    fprintf (F, "add_row   %8.5lf %8.5lf %8.5lf %8.5lf\n",
-		     matr->_w, matr->_x, matr->_y, matr->_z); 
-	}
-	break;
-    case 0x20:
-	fprintf (F, "%s\n%s\n", mhdr20a, mhdr20b);
-	for (i = 0; i < _dec.nspkr; i++, matr++)
-	{
-	    fprintf (F, "add_row   %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf\n",
-		     matr->_w, matr->_x, matr->_y, matr->_u, matr->_v); 
-	}
-	break;
-    case 0x21:
-	fprintf (F, "%s\n%s\n", mhdr21a, mhdr21b);
-	for (i = 0; i < _dec.nspkr; i++, matr++)
-	{
-	    fprintf (F, "add_row   %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf\n",
-		     matr->_w, matr->_x, matr->_y, matr->_z, matr->_u, matr->_v); 
-	}
-	break;
-    case 0x22:
-	fprintf (F, "%s\n%s\n", mhdr22a, mhdr22b);
-	for (i = 0; i < _dec.nspkr; i++, matr++)
-	{
-	    fprintf (F, "add_row   %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf\n",
-		     matr->_w, matr->_x, matr->_y, matr->_z, matr->_r, matr->_s, matr->_t, matr->_u, matr->_v); 
-	}
-	break;
-    case 0x30:
-	fprintf (F, "%s\n%s\n", mhdr30a, mhdr30b);
-	for (i = 0; i < _dec.nspkr; i++, matr++)
-	{
-	    fprintf (F, "add_row   %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf\n",
-		     matr->_w, matr->_x, matr->_y, matr->_u, matr->_v, matr->_p, matr->_q); 
-	}
-	break;
-    case 0x31:
-	fprintf (F, "%s\n%s\n", mhdr31a, mhdr31b);
-	for (i = 0; i < _dec.nspkr; i++, matr++)
-	{
-	    fprintf (F, "add_row   %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf %8.5lf\n",
-		     matr->_w, matr->_x, matr->_y, matr->_z, matr->_u, matr->_v, matr->_p, matr->_q); 
-	}
+    case FUMA:
+	gc = gc_fuma;
 	break;
+    default:
+	gc = 0;
     }
 
+    fprintf (F, "/%s/{\n", pref);
+    fprintf (F, "order_gain    %8.5lf %8.5lf %8.5lf %8.5lf\n", gain [0], gain [1], gain [2], gain [3]);
+
+    for (j = 0, M = matr; j < _dec.nspkr; j++, M++)
+    {
+        fprintf (F, "add_row   ");
+	for (i = 0, m = _dec.cmask; i < 16; i++, m >>= 1)
+	{
+	    if (m & 1) fprintf (F, " %9.6lf", M->_coeff [i] * gc [i]);
+	}
+	fprintf (F, "\n");
+    }
     fprintf (F, "/}\n\n");
 }
 
@@ -286,7 +367,7 @@ int AD_conf::load (const char *file)
     char          *p, *q;
     char          buff [1024];
     int           line, stat, n, r;
-    int           i1, nsp, nlf, nhf, dim, ord, *cnt;
+    int           i1, nsp, nlf, nhf, *cnt;
     float         f1, f2, f3;
     char          s1 [64];
 
@@ -302,8 +383,6 @@ int AD_conf::load (const char *file)
     reset ();
     stat = 0;
     line = 0;
-    dim = 0;
-    ord = 0;
     nsp = 0;
     nlf = 0;
     nhf = 0;
@@ -331,7 +410,7 @@ int AD_conf::load (const char *file)
 	    else
 	    {
 		q += n;
-		if ((i1 < 1) && (i1 > 2))
+		if ((i1 < 1) && (i1 > 3))
 		{
 		    fprintf (stderr, "Line %d: unknown version %d.\n", line, i1);
 		    stat = ERROR;
@@ -348,7 +427,8 @@ int AD_conf::load (const char *file)
         else if (!strcmp (p, "/dec/hor_order"))
 	{
 	    if (R || S) stat = SCOPE;
-	    if (sscanf (q, "%d%n", &i1, &n) != 1) stat = ARGS;
+	    else if (_fform > 2) stat = COMM;
+	    else if (sscanf (q, "%d%n", &i1, &n) != 1) stat = ARGS;
 	    else
 	    {
 		q += n;
@@ -357,68 +437,35 @@ int AD_conf::load (const char *file)
 		    fprintf (stderr, "Line %d: illegal horizontal order %d.\n", line, i1);
 		    stat = ERROR;
 		}
-		else _dec.h_ord = i1;
+		else _h_ord = i1;
 	    }
 	}
         else if (!strcmp (p, "/dec/ver_order"))
 	{
 	    if (R || S) stat = SCOPE;
-	    if (sscanf (q, "%d%n", &i1, &n) != 1) stat = ARGS;
+	    else if (_fform > 2) stat = COMM;
+	    else if (sscanf (q, "%d%n", &i1, &n) != 1) stat = ARGS;
 	    else
 	    {
-		q += n;
-		if ((i1 < 0) || (i1 > _dec.h_ord) || (i1 + _dec.h_ord > 4)) // HACK
+ 	        q += n;
+  	        if ((i1 < 0) || (i1 > _h_ord) || (i1 + _h_ord > 4))
 		{
 		    fprintf (stderr, "Line %d: illegal vertical order %d.\n", line, i1);
 		    stat = ERROR;
 		}
-		else _dec.v_ord = i1;
+                else _v_ord = i1;
 	    }
+
 	}
-        else if (!strcmp (p, "/dec/dimension"))
+        else if (!strcmp (p, "/dec/chan_mask"))
 	{
 	    if (R || S) stat = SCOPE;
-	    if (sscanf (q, "%d%n", &i1, &n) != 1) stat = ARGS;
+	    else if (_fform != 3) stat = COMM;
+	    else if (sscanf (q, "%x%n", &i1, &n) != 1) stat = ARGS;
 	    else
 	    {
-		q += n;
-		if ((i1 != 2) && (i1 != 3))
-		{
-		    fprintf (stderr, "Line %d: illegal decoder dimension %d.\n", line, i1);
-		    stat = ERROR;
-		}
-		else
-		{
-                    dim = i1;
-                    if (ord)
-                    {
-	                _dec.h_ord = ord;
-	                _dec.v_ord = (dim == 3) ? ord : 0;
-		    }
-		}
-	    }
-	}
-        else if (!strcmp (p, "/dec/amb_order"))
-	{
-	    if (R || S) stat = SCOPE;
-	    if (sscanf (q, "%d%n", &i1, &n) != 1) stat = ARGS;
-	    else
-	    {
-		q += n;
-		if ((i1 != 1) && (i1 != 2))
-		{
-		    fprintf (stderr, "Line %d: illegal decoder order %d.\n", line, i1);
-		    stat = ERROR;
-		}
-		else
-		{
-                    ord = i1;
-                    if (dim)
-                    {
-	                _dec.h_ord = ord;
-	                _dec.v_ord = (dim == 3) ? ord : 0;
-		    }
-		}
+	       q += n;
+               _dec.cmask = i1;
 	    }
 	}
         else if (!strcmp (p, "/dec/speakers"))
@@ -602,12 +649,9 @@ int AD_conf::load (const char *file)
 		    S->_r = f1;
 		    S->_a = f2;
 		    S->_e = f3;
-		    if (sscanf (q, "%s%n", s1, &n) == 1)
-		    {
-			strcpy (S->_jackp, s1);
-			q += n;
-		    }
-		    S++;
+		    n = sstring (q, S->_jackp, 64);
+		    q += n;
+  	            S++;
 		    nsp++;
 		}
 	    }
@@ -630,6 +674,7 @@ int AD_conf::load (const char *file)
 		R = _lfmatrix;
 		G = _lfgain;
 		cnt = &nlf;
+		if ((_fform < 3) && (_chmap == 0)) makechmap ();
 	    }
 	}
 	else if (!strcmp (p, "/lfmatrix/{"))
@@ -650,6 +695,7 @@ int AD_conf::load (const char *file)
 		R = _lfmatrix;
 		G = _lfgain;
 		cnt = &nlf;
+		if ((_fform < 3) && (_chmap == 0)) makechmap ();
 	    }
 	}
         else if (!strcmp (p, "/hfmatrix/{"))
@@ -670,12 +716,20 @@ int AD_conf::load (const char *file)
 		R = _hfmatrix;
 		G = _hfgain;
 		cnt = &nhf;
+		if ((_fform < 3) && (_chmap == 0)) makechmap ();
 	    }
 	}
         else if (!strcmp (p, "order_gain"))
 	{
-	    switch (_dec.h_ord) 
+	    if (_fform == 3)
 	    {
+	        if (sscanf (q, "%f%f%f%f%n\n", G + 0, G + 1, G + 2, G + 3, &n) != 4) stat = ARGS;
+ 	        else q += n;
+	    }
+	    else
+	    {
+  	        switch (_h_ord) 
+	        {
 	    case 1:
 	        if (sscanf (q, "%f%f%n\n", G + 0, G + 1, &n) != 2) stat = ARGS;
  	        else q += n;
@@ -688,6 +742,7 @@ int AD_conf::load (const char *file)
 	        if (sscanf (q, "%f%f%f%f%n\n", G + 0, G + 1, G + 2, G + 3, &n) != 4) stat = ARGS;
  	        else q += n;
 		break;
+	        }
 	    }
 	}
         else if (!strcmp (p, "add_row"))
@@ -700,7 +755,8 @@ int AD_conf::load (const char *file)
 	    }
             else
 	    {
-		stat = read_matrow (&q, R);
+		if (_fform < 3) stat = read_matrow2 (&q, R);
+		else            stat = read_matrow3 (&q, R);
 	        if (!stat)
 	        {
 		    R++;
@@ -748,49 +804,81 @@ int AD_conf::load (const char *file)
 }
 
 
-int AD_conf::read_matrow (char **q, Matrow *R)
+void AD_conf::makechmap (void)
 {
-    int n, r;
-
-    switch ((_dec.h_ord << 4) | _dec.v_ord)
+    switch ((_h_ord << 4) | _v_ord)
     {
     case 0x10:
-	r = sscanf (*q, "%f%f%f%n",
-                    &R->_w, &R->_x, &R->_y, &n); 
-	if (r != 3) return ARGS;
-	break;
+	_chmap = chmap10;
+	_dec.cmask = 0x000B;
+        break;
     case 0x11:
-	r = sscanf (*q, "%f%f%f%f%n",
-                    &R->_w, &R->_x, &R->_y, &R->_z, &n); 
-	if (r != 4) return ARGS;
-	break;
+	_chmap = chmap11;
+	_dec.cmask = 0x000F;
+        break;
     case 0x20:
-	r = sscanf (*q, "%f%f%f%f%f%n",
-		    &R->_w, &R->_x, &R->_y, &R->_u, &R->_v, &n); 
-	if (r != 5) return ARGS;
-	break;
+	_chmap = chmap20;
+	_dec.cmask = 0x011B;
+        break;
     case 0x21:
-	r = sscanf (*q, "%f%f%f%f%f%f%n",
-		    &R->_w, &R->_x, &R->_y, &R->_z, &R->_u, &R->_v, &n); 
-	if (r != 6) return ARGS;
-	break;
+	_chmap = chmap21;
+	_dec.cmask = 0x011F;
+        break;
     case 0x22:
-	r = sscanf (*q, "%f%f%f%f%f%f%f%f%f%n",
-		    &R->_w, &R->_x, &R->_y, &R->_z, &R->_r, &R->_s, &R->_t, &R->_u, &R->_v, &n); 
-	if (r != 9) return ARGS;
-	break;
+	_chmap = chmap22;
+	_dec.cmask = 0x01FF;
+        break;
     case 0x30:
-	r = sscanf (*q, "%f%f%f%f%f%f%f%n",
-		    &R->_w, &R->_x, &R->_y, &R->_u, &R->_v, &R->_p, &R->_q, &n); 
-	if (r != 7) return ARGS;
-	break;
+	_chmap = chmap30;
+	_dec.cmask = 0x831B;
+        break;
     case 0x31:
-	r = sscanf (*q, "%f%f%f%f%f%f%f%f%n",
-		    &R->_w, &R->_x, &R->_y, &R->_z, &R->_u, &R->_v, &R->_p, &R->_q, &n); 
-	if (r != 8) return ARGS;
-	break;
+	_chmap = chmap31;
+	_dec.cmask = 0x831F;
+        break;
     }
-    *q += n;
+}
+
+
+int AD_conf::read_matrow2 (char **q, Matrow *R)
+{
+    int     i, j, n, r;
+    float   v;
+
+    R->reset ();
+    for (i = 0; i < 16; i++)
+    {
+	j = _chmap [i];
+	if (j < 0) break;
+        r = sscanf (*q, "%f%n", &v, &n);
+        *q += n;
+        if (r) R->_coeff [j] = v;
+        else return ARGS;
+    }
+    if (_dec.scale == SN3D) for (i = 0; i < 16; i++) R->_coeff [i] /= gc_sn3d [i];
+    if (_dec.scale == FUMA) for (i = 0; i < 16; i++) R->_coeff [i] /= gc_fuma [i];
+    return 0;
+} 
+
+
+int AD_conf::read_matrow3 (char **q, Matrow *R)
+{
+    int     i, m, n, r;
+    float   v;
+
+    R->reset ();
+    for (i = 0, m = _dec.cmask; m; i++, m >>= 1)
+    {
+	if (m & 1)
+	{
+	    r = sscanf (*q, "%f%n", &v, &n);
+            *q += n;
+	    if (r) R->_coeff [i] = v;
+	    else return ARGS;
+	}
+    }
+    if (_dec.scale == SN3D) for (i = 0; i < 16; i++) R->_coeff [i] /= gc_sn3d [i];
+    if (_dec.scale == FUMA) for (i = 0; i < 16; i++) R->_coeff [i] /= gc_fuma [i];
     return 0;
 } 
 
