@@ -1,6 +1,6 @@
 // ----------------------------------------------------------------------------
 //
-//  Copyright (C) 2006-2010 Fons Adriaensen <fons@kokkinizita.net>
+//  Copyright (C) 2006-2011 Fons Adriaensen <fons@linuxaudio.org>
 //    
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -37,7 +37,7 @@ public:
 
     void update (float v);
 
-    static void load_images (X_display *disp, X_scale_style *scale);
+    static int load_images (X_display *disp, X_scale_style *scale);
 
 private:
 
